type Event_T = [name:string, payload:any] | ['entry', any];
export interface Machine_T<S,E extends Event_T,C> {
  states: Array<State_T<S,E,C>>
}
export interface State_T<S,E extends Event_T,C> {
  name: S;
  ons: Array<On_T<S,E,C>>;
}
export interface On_T<S,E extends Event_T,C> {
  eventName: E[0];
  reactions: Array<Do_T<S,E,C> | Goto_T<S>>;
};
export interface Do_T<S,E extends Event_T, C> {
  type: 'Do';
  fn: DoFn_T<S,E,C>;
};
export type DoFn_T<S,E extends Event_T, C> = (ctx:C,e:E,self:Interpreter_T<S,E,C>)=>any;
export interface Goto_T<S> {
  type: 'Goto';
  targetStateName: S;
};

export const Machine = function<S,E extends Event_T,C>(...states:Array<State_T<S,E,C>>) : Machine_T<S,E,C> { return {states}; };
export const State = function<S,E extends Event_T,C>(name:S, ...ons:Array<On_T<S,E,C>>) : State_T<S,E,C>{ return {name, ons}; };
export const On = function<S,E extends Event_T,C>(eventName:E[0], ...reactions:Array<Do_T<S,E,C> | Goto_T<S>>) : On_T<S,E,C>{ return {eventName, reactions}; };
export const Do = function<S,E extends Event_T, C>(fn:DoFn_T<S,E,C>) : Do_T<S,E,C>{ return {type:'Do', fn}; };
export const Goto = function<S>(targetStateName:S) : Goto_T<S> { return {type:'Goto', targetStateName} };

interface Tick_T<S,E extends Event_T, C> {
  doFunctions: Array<DoFn_T<S,E,C>>
};
const Tick = function<S,E extends Event_T,C>(doFunctions : Array<DoFn_T<S,E,C>>) : Tick_T<S,E,C>{ return {doFunctions}; };


export interface Interpreter_T<S,E extends Event_T,C> {
  machine: Machine_T<S,E,C>;
  state: S;
  context: C;
  tickQueues:Array<TickQueue_T<S,E,C>>
}
type TickQueue_T<S,E extends Event_T,C> = Array<Tick_T<S,E,C>>;
export function interpret<S,E extends Event_T,C>(machine:Machine_T<S,E,C>, options:{state?:S, context:C}) : Interpreter_T<S,E,C>{
  let {state, context} = options;
  if(typeof state === 'undefined'){ state = machine.states[0].name; }
  const interpreter = {machine, state, context, tickQueues:[]}
  //@ts-ignore
  send(interpreter, ['entry', null] );
  return interpreter;
}

/**  Helper function for `send()`
 */
function getState<S,E extends Event_T,C>(interpreter : Interpreter_T<S,E,C>) : State_T<S,E,C>{
  return interpreter.machine.states.find((state)=>state.name===interpreter.state) as unknown as State_T<S,E,C>;
}
/**  Helper function for `send()`
 */
function getOns<S,E extends Event_T,C>(state : State_T<S,E,C>, event:E) : Array<On_T<S,E,C>>{
  return state.ons.filter((on)=>on.eventName===event[0]);
}
/**  Helper function for `send()`
 */
function noop(){}
/**  Inject an Event into the Interpreter's "tick queue".
 * 
 * An event can be signify something "new" happening, such that its reactions should run on the next Tick;
 * or it can signify a milestone "within" the current Tick, such that a Tick can be thought of as having 
 * "sub-Ticks".
 * 
 * This distinction is significant for proper ordering of reaction execution, and also for determining
 * whether to run a reaction at all. If an Event is received, and is specified to be applied on a past 
 * Tick, it is discarded.
 */
export function send<S,E extends Event_T,C>(interpreter : Interpreter_T<S,E,C>, event:E){
  const state = getState(interpreter);
  const onsTree = getOns(state, event);
  const reactions = onsTree
    .map((on)=>on.reactions)
    .flat();
  const indexOfFirstGoto = reactions.findIndex((reaction)=>reaction.type==='Goto');
  const indexOfFinalReaction = indexOfFirstGoto === -1 ? reactions.length-1 : indexOfFirstGoto;
  const reactionsUntilFirstGoto = reactions.slice(0, indexOfFinalReaction+1);
  const functionsToRunInTick = reactionsUntilFirstGoto
    .map((reaction)=>{
      if(reaction.type === 'Do'){
        return reaction.fn;
      }
      else if(reaction.type === 'Goto'){
        return (ctx:C,e:E,self:Interpreter_T<S,E,C>)=>{
          self.state = reaction.targetStateName;
          //@ts-ignore
          send(self, ['entry', e] );
        };
      }
      else{
        return noop;
      }
    });
  const tick = Tick(functionsToRunInTick);
  tick.doFunctions.forEach((fn)=>{ fn(interpreter.context, event, interpreter); });
}

export const Spawn = function(){};
export const Unspawn = function(){};